<?php
/**
 * The database_object is an object database mapper for mysql and AdoDB (e.g.
 * Postre SQL)
 *
 * This class provides a layer to access a specified table of the database. 
 * Name of the table is specified by the name of the subclass. 
 *
 * Attributes are automatically determined from database and stored in the 
 * instance.
 *
 * @author Jonas Mitschang, jonen@mitschang.net
 * http://mitschang.net/software/dbo.html
 *
 * Version: 1.21
 * Changelog:
 * 1.21 
 *   - db_adodb added
 *   - mysql adapter can be initialized with an existing connection
 *   - bugfix when using prefix (e.g. joomla jos_...)
 * 1.2
 *   - first public release
 */

class database_object {
    /**
     * Automatically save changes off attributes to the database? 
     */
    public $autosave = false;

    /** 
     * Oject exists in database
     */
    public $existent = false;

    /**
     * Column definitions 
     * e.g. var $coldef = array('user0_id' => 'user', 'user1_id' => 'user);
     */
    protected $coldef = array();

    //! @cond
    /** 
     * Hash with all attributes, that this database object contains in the 
     * database. The hash has the following format: 
     *   name => true (at the moment, TODO later name => datatype)
     */
    protected $_attributes;

    /** 
     * Hash with all attributes, that this database object contains in the
     * database. The hash has the following format: 
     *   attribute name => value
     */
    protected $_values = array();

    /** Hash with all (loaded) sub-objects
     */
    private $_subobj;

    /** modified fields are marked in here */
    protected $modified;

    /** 
     * The name of this database object and the corresponding table.
     * Do not make this static - will not work because all database_objects 
     * would have the same name in this case.
     */
    protected $table_name = null;

    /**
     * Static cache for attributes and objects 
     */
    static protected $cache = array();

    /**
     * database connection
     */
    static private $db = null;
    //! @endcond

    /**
     * Constructor
     *
     * @param $id
     *   If id given: fetch the corresponding row from database
     *     Example: $vehicle = new vehicle(10); // fetch vehicle where id=10
     *   If not: create an empty object, that may be saved lateron (INSERT)
     *     Example: $new_v = new vehicle(); $new_v->color = 'red'; $new_v->save();
     * @param $load_attr
     *   Is only used internally and should never be specified
     */
    function database_object($id = null, $load_attr = true) {
        // db is checked here once - so no further checking needed
        assert('isset(self::$db)'); 

        // which table to use?
        $this->table_name = self::$db->prefix.get_class($this);

        // get the attributes
        $this->_attributes = self::get_cache(get_class($this), '_attributes');
        if (!isset($this->_attributes)) {
            // attributes not cached => fetch from database
            $this->_attributes = self::$db->fetch_columns($this->table_name);

            // and store to cache
            self::set_cache(get_class($this), '_attributes', $this->_attributes);
        }
        assert('array_key_exists("id", $this->_attributes)');

        // only load if a valid $id is given
        $this->set_with_type('id', 0);
        if (!isset($id) || ($id === null)) return;

        assert('is_numeric($id)');
        if (!$load_attr) {
            // do not load attrs. used for get_all
            $this->id = $id;
            // object exists in database
            $this->existent = true;
            return;
        }

        // check cache and load from cache if existing
        $cached = self::get_cache(get_class($this), $id);
        if (isset($cached) && ($cached != null)) {
            foreach ($cached as $k => $v) $this->$k = $v;
            return;
        } 

        // load all database attrobutes
        if ($this->load_attributes($id)) $this->existent = true;
        self::set_cache(get_class($this), $id, $this);
    }

    public static function init($db) {
        assert('$db instanceof db_adapter');
        self::$db = $db;

        // init database adapter
        self::$db->init();
    }

    public function get_attributes() {
        assert('is_array($this->_attributes)');
        return $this->_attributes;
    }

    /**
     * Write and read the cache
     */
    static private function get_cache($name, $var) {
        if (self::$cache == null) return null;
        $name = $name.'__'.$var;
        if (!array_key_exists($name, self::$cache)) return null;
        return self::$cache[$name];
    }
    static private function set_cache($name, $var, $val) {
        self::$cache[$name.'__'.$var] = $val;
    }

    /**
     * Get the name (the table name) of this database object
     */
    final public function get_table_name() {
        return $this->table_name;
    }

    // Build the SQL-Query to load all attributes from Database
    private function select_statement($where = null, $limit = null, $order = null) {
        // let _get_all_query handle the rest
        $query = 'SELECT * FROM '.$this->table_name;
        $query .= self::$db->get_where($where);
        $query .= self::$db->get_order($order);
        $query .= self::$db->get_limit($limit);
        return $query;
    }

    // Not documented => not in doxy. See get_all
    function _get_all($where = null, $limit = null, $order = null) {
        return $this->_get_all_query($this->select_statement($where, $limit, $order));
    }

    function _count($where = null) {
        $query = 'SELECT COUNT(*) AS count FROM '.$this->table_name.
            self::$db->get_where($where).';';
        return (int)self::select_single($query);
    }

    static function select_single($query) {
        return self::$db->select_single($query);
    }

    // Not documented => not in doxy. See get_all_query
    function _get_all_query($query) {
        assert('is_string($query)');

        // create objects
        $all = self::$db->query_all($query);
        $result = array(); 
        $obj_name = get_class($this);
        foreach ($all as $item) {
            // cached?
            $obj = self::get_cache($obj_name, $item['id']);
            if (!isset($obj) || ($obj == null)) {
                // not cached: create new object
                $obj = new $obj_name($item['id'], false);
                foreach ($this->_attributes as $attr => $type) {
                    $obj->set_with_type($attr, $item[$attr]);
                }

                // save cache
                self::set_cache(get_class($this), $obj->id, $obj);
            }

            // save to result array
            array_push($result, $obj);
        }
        return $result;
    }

    /**
     * Check if the given attribute exists
     */
    function attribute_exists($attr) {
        return isset($this->_attributes[$attr]);
    }

    /**
     * Return the type of an attribute
     */
    function attribute_type($attr) {
        assert('isset($this->_attributes[$attr])');
        return $this->_attributes[$attr];
    }

    private function load_owners($what=null, $order='id ASC', $force=false) {
        if (!isset($what)) {
            // load all
            die('loading all owners is not implemented yet - and never will');
        }

        if (is_array($what)) {
            // load multiple attributes
            foreach ($what as $attr) $this->load_owners($attr);
        } else {
            // load one owner
            $owner = new $what();

            // check if we really belong to owner
            $our_id = get_class($this)."_id";
            if (!$owner->attribute_exists($our_id)) 
                die("Cannot load owners: ".$what." has no attribute ".$our_id);

            // already loaded? noting to do
            if (!$force && isset($this->_subobj[$what])) return;

            // load (ordered by id per default)
            if ($this->id == 0) {
                $this->_subobj[$what] = array();
            } else {
                $this->_subobj[$what] = $owner->_get_all(
                    array(get_class($this) => $this), null, $order);
            }
        }
    }

    /**
     * Loads objects from a db-table to a local field
     *
     * Example:
     *   $o->load('vehicle') uses the column vehicle_id to load the
     *   vehicle with that id from the vehicle table. The resulting
     *   vehicle object is stored to $obj->vehicle
     *   Resulting query: SELECT * FROM vehicle WHERE id=$o->vehicle_id
     */
    private function load($what) {
        if (is_array($what)) {
            // load multiple attributes
            foreach ($what as $attr) $this->load($attr);
            return;
        } 

        // load one attribute
        $what_id = $what . "_id";
        if (isset($this->coldef[$what])) {
            $objname = $this->coldef[$what];
        } else {
            $objname = $what;
        }
        assert('$this->attribute_exists($what_id)');

        // id given?
        $id = $this->_values[$what_id];
        if (!isset($id)) {
            // no id => nothing to load => unset
            $this->_subobj[$what] = new $what();
            return;
        }

        // already loaded (the same)? noting to do
        if (isset($this->_subobj[$what]) && ($id == $this->_subobj[$what]->id)) return;

        // got id => laod from db
        // shortcut: try to fetch from cache
        $new_obj = self::get_cache($what, $id);
        if (!isset($new_obj)) $new_obj = new $objname($id);
        $this->_subobj[$what] = $new_obj;
    }

    public function attributes_type($attr) {
        assert('isset($this->_attributes[$attr])');
        return $this->_attributes[$attr];
    }

    private function set_with_type($attr, $value) {
        assert('isset($this->_attributes[$attr])');
        if ($this->attributes_type($attr) === 'bool') {
            $value = $value == 't';
        }
        $this->_values[$attr] = $value;
    }

    //! @cond
    /**
     * Method load
     * This method loads all attributes from a db-table. The table's name will
     * be extracted from the name of the subclass.
     */
    protected function load_attributes($id) {
        // select from SQL server
        $query = $this->select_statement('id='.$id);
        $record = self::$db->query_row($query);

        // not found in database
        if (!$record) return false;

        // found
        foreach ($this->_attributes as $attr => $type) {
            $this->set_with_type($attr, $record[$attr]);
        }
        assert('$this->id == $id');
        return true;
    }
    //! @endcond

    /**
     * Saves the database-object into the database.
     * If existent, then call the update-method, if not, add a new entry
     */
    public function save() {
        if ($this->existent) {
            $this->update();
        } else  {
            $this->insert();
        }

        // now the object must be existent
        assert('$this->existent');

        // also update the cache if the object is accesed later on
        //echo "ID: ".$this->id;
        // FIXME: nicht ganzen cache loeschen nur das element, aber geht 
        // irgendwie nicht
        // self::set_cache($this->table_name, $this->id, null /*$this*/);
        self::cache_clear();
    }

    /**
     * Clear the cache. Should normally not be used - just in case something 
     * behaves 'strange'
     */
    static public function cache_clear() {
        self::$cache = array();
    }

    /**
     * Disable the cache.
     * Can be used if plenty of objects will be kelled and thus the php memory 
     * limit would be hit.
     */
    static public function cache_disable() {
        self::$cache = null;
    }

    //! @cond
    /**
     * Method update
     * 
     * Updates the representing row in the database with current attributes of
     * the database-object
     */
    protected function update() {
        assert('$this->existent /* trying to update inexistent database object */');
        assert('$this->id > 0');

        // no changes => nothing to do
        if (!isset($this->modified)) return;
        assert('is_array($this->modified)');
        if (sizeof($this->modified) == 0) return;

        // what to update?
        $record = array();
        foreach ($this->modified as $attr => $mod) {
            if ($mod) $record[$attr] = $this->$attr; 
        }
        self::$db->update($this->table_name, $record, $this->id);

        // okay no modifications now
        $this->modified = array();
    }
    //! @endcond

    /**
     * Optimize this table. Call e.g. from Admin interface or once on long 
     * program run.
     */
    public static function optimize() {
        self::$db->optimize($this->table_name);
    }

    /**
     * Delete the representing row in database 
     */
    public function delete() {
        assert('$this->existent /* ERROR: trying to delete inexistent database object */');
        assert('$this->id > 0');

        $query = 'DELETE FROM '.$this->table_name.' WHERE id='.$this->id;

        // TODO evaluate result
        self::$db->query($query);

        // also remove from cache
        self::set_cache(get_class($this), $this->id, null);

        // okay, no longer in database
        $this->id = null;
        $this->existent = false;
    }

    //! @cond
    /**
     * Adds a new row to database with current attributes
     * of the database-object
     */
    protected function insert() {
        $record = array();
        foreach ($this->_attributes as $attr => $type) {
            $value = $this->$attr;
            if (($attr == 'id') && ($value === 0)) continue; // id=0 =>autoindex
            if (!isset($value)) continue; // not set => default
            $record[$attr] = $value;
        }

        // okay created => store id
        $this->id = self::$db->insert($this->table_name, $record);

        // okay no modifications now
        $this->modified = null;
        $this->existent = true;
    }
    //! @endcond

    // Magic methods
    public function __get($k) {
        if ($this->attribute_exists($k)) {
            // return the field's value
            if (!isset($this->_values[$k])) return null;
            return $this->_values[$k];
        } else if (isset($this->_subobj[$k])) {
            return $this->_subobj[$k];
        }

        // load sub object attribute
        $k_id = $k . "_id";
        if ($this->attribute_exists($k_id)) {
            $this->load($k);
            return $this->_subobj[$k];
        } 

        // not found: has_many?
        if (isset($this->_subobj[$k])) return $this->_subobj[$k];
        if (isset($this->has_many)) $this->load_owners($this->has_many);
        if (isset($this->_subobj[$k])) return $this->_subobj[$k];

        // not found
        die("Property $k not found in the database_object ".
            $this->table_name." (reading). Add column $k to table ".
            $this->table_name."\n");
    }
    public function __set($k, $v) {
        if ($this->attribute_exists($k)) {
            // update the modified field
            if (!isset($this->_values[$k]) || ($this->_values[$k] !== $v)) {
                $this->modified[$k] = true; 

                // update the field itself and save if desired
                $this->_values[$k] = $v;
                if ($this->autosave && $this->existent) $this->save();
            }

            // return
            return $this->_values[$k];
        } 

        // load one attribute
        $k_id = $k . "_id";
        if ($this->attribute_exists($k_id)) {
            if ($v === null) {
                // null is stored as _id = 0
                $this->$k_id = 0;
            } else {
                // update _id field
                assert('$v instanceof database_object');
                $this->$k_id = $v->id;
            }

            // object was loaded? reload
            if (isset($this->_subobj[$k])) {
                $this->load($k);
            }
            return $this->_subobj[$k];
        } 

        // not found
        die("Property $k not found in the database_object ".
            $this->table_name." (writing). Add column $k to table ".
            $this->table_name."\n");
    }
    public function __isset($k) {
        return isset($this->_values[$k]);
    }
    public function __unset($k) {
        unset($this->_values[$k]);
    }

    /**
     * Is the given attribute changed (not yet saved to the database)
     */
    public function is_modified($attr) {
        assert('is_string($attr)');
        if ($this->modified == null) return false;
        if (!isset($this->modified[$attr])) return false;
        return $this->modified[$attr] != false;
    }

    /**
     * Load a bunch of database objects from the corresponding table.
     *
     * @param $where
     *   The SQL 'WHERE' part of the select statement. Set to null or 'TRUE' to 
     *   select all objects. Use array notation for better readabilty.
     *   Example: Get all red vehicles that use diesel:
     *     vehicle::get_all(array(
     *       'color' => 'red', 
     *       'fuel' => 'diesel'
     *     ));
     *
     * @param $limit
     *   The SQL 'LIMIT' part of the select statement. Set to null to get all 
     *   objects with no limit. 
     *   The first row is identified (like normal for SQL) with 'start'=>0.
     *
     * @param $order
     *   The SQL 'ORDER' part of the select statement. Set to null to irgnore
     *   ordering.
     *
     * @return an array of databse objects foo. The array may be empty if there 
     *   are no matching objects.
     *
     * Example: 
     *   Get first ten employees, ordered by login name:
     *   $all = employee::get_all(
     *     null, array('start'=>0, 'count'=>10), 'login ASC');
     */
    public static function get_all($where = null, $limit=null, $order=null) {
        // has to be called from static context
        $obj_name = get_called_class();
        $obj = new $obj_name();
        return $obj->_get_all($where, $limit, $order);
    }

    /**
     * Return the last-created object (this is by definition the object with the
     * highest id value)
     *
     * @param $where
     *   If where is set, then the last object matching the specified WHERE 
     *   condition will be returned.
     *
     * @returns The last database-object ih this table. If there is no matching 
     *   object, the function will return null.
     */
    public static function get_last($where = null) {
        return self::get_single($where, array('start'=>0, 'count'=>1), "id DESC");
    }

    /**
     * Return an array of objects that result from the given query
     * Accordingly SELECTs over multiple tables can be used
     */
    public static function get_by_query($query, $limit = null) {
        assert('is_string($query)');
        $obj_name = get_called_class();
        $obj = new $obj_name();
        return $obj->_get_all_query($query, $limit);
    }

    /**
     * Return the only object that matches the expression
     * The code checks, that there is only one match.
     * If there is no match, null is returned.
     */
    public static function get_single($where=null, $limit=null, $order=null) {
        $all = self::get_all($where, $limit, $order);
        assert('is_array($all)');
        assert('sizeof($all) <= 1 /* size='.sizeof($all).' */');
        if (sizeof($all) == 0) return null;
        return $all[0];
    }

    /**
     * Count the number of objects that match the given condition.
     * @see get_all to learn more abit the conditions ($where)
     *
     * @param where Condition like described in get_all
     * @returns The number of matches as integer
     */
    public static function count($where = null) {
        // has to be called from static context
        $obj_name = get_called_class();
        $obj = new $obj_name();
        return $obj->_count($where);
    }

    /**
     * Count number of rows that would be returned by the given query
     */
    public static function count_query($query = null) {
        // TOOD assert: ohne LIMIT
        $query = preg_replace('/SELECT .*? FROM/', 'SELECT COUNT(*) FROM', $query, 1);
        $query = preg_replace('/ORDER BY.*/', '', $query, 1);
        return (int)self::select_single($query);
    }
}

/**
 * Retro-support of get_called_class()
 * (PHP 5 >= 5.3.0)
 * get_called_class — the "Late Static Binding" class name
 */
if (!function_exists('get_called_class')) {
    function get_called_class($bt = false,$l = 1) {
        if (!$bt) $bt = debug_backtrace();
        if (!isset($bt[$l])) {
            throw new Exception(
                "Cannot find called class -> stack level too deep.");
        }
        if (!isset($bt[$l]['type'])) {
            throw new Exception ('type not set');
        }
        switch ($bt[$l]['type']) {
        case '::':
            $lines = file($bt[$l]['file']);
            $i = 0;
            $callerLine = '';
            do {
                $i++; 
                $callerLine = $lines[$bt[$l]['line']-$i] . $callerLine;
            } while (stripos($callerLine,$bt[$l]['function']) === false);
            preg_match('/([a-zA-Z0-9\_]+)::'.$bt[$l]['function'].'/',
                $callerLine,
                $matches);
            if (!isset($matches[1])) {
                // must be an edge case.
                throw new Exception ("Could not find caller class: originating method call is obscured.");
            }
            switch ($matches[1]) {
            case 'self':
            case 'parent':
                return get_called_class($bt,$l+1);
            default:
                return $matches[1];
            }
            // won't get here.
        case '->': switch ($bt[$l]['function']) {
            case '__get':
                // edge case -> get class of calling object
                if (!is_object($bt[$l]['object'])) throw new Exception ("Edge case fail. __get called on non object.");
                return get_class($bt[$l]['object']);
            default: return $bt[$l]['class'];
        }

        default: throw new Exception ("Unknown backtrace method type");
        }
    }
}

